pub mod prime_utils {
    use lazy_static::lazy_static;
    use num::{BigUint, One, Zero};

    use crate::primality_test::primality_tests::is_probably_prime;

    #[must_use]
    pub fn log_2(x: &BigUint) -> u64 {
        x.bits() - 1
    }

    #[must_use]
    pub fn is_prime_default(number: &BigUint) -> bool {
        lazy_static! {
            static ref DEFAULTVEC: Vec<BigUint> = {
                let mut vec = Vec::new();
                vec.push(BigUint::from(2u8));
                vec.push(BigUint::from(3u8));
                vec.push(BigUint::from(5u8));
                vec.push(BigUint::from(7u8));
                vec.push(BigUint::from(11u8));
                vec.push(BigUint::from(13u8));
                vec.push(BigUint::from(17u8));
                vec.push(BigUint::from(19u8));
                vec.push(BigUint::from(23u8));
                vec.push(BigUint::from(29u8));
                vec.push(BigUint::from(31u8));
                vec.push(BigUint::from(37u8));
                vec.push(BigUint::from(41u8));
                vec.push(BigUint::from(43u8));
                vec.push(BigUint::from(47u8));
                vec.push(BigUint::from(53u8));
                vec.push(BigUint::from(59u8));
                vec.push(BigUint::from(61u8));
                vec.push(BigUint::from(67u8));
                vec.push(BigUint::from(71u8));
                vec.push(BigUint::from(73u8));
                vec.push(BigUint::from(79u8));
                vec.push(BigUint::from(83u8));
                vec.push(BigUint::from(89u8));
                vec.push(BigUint::from(97u8));
                vec.push(BigUint::from(101u8));
                vec.push(BigUint::from(103u8));
                vec.push(BigUint::from(107u8));
                vec.push(BigUint::from(109u8));
                vec.push(BigUint::from(113u8));
                vec.push(BigUint::from(127u8));
                vec.push(BigUint::from(131u8));
                vec.push(BigUint::from(137u8));
                vec.push(BigUint::from(139u8));
                vec.push(BigUint::from(149u8));
                vec.push(BigUint::from(151u8));
                vec.push(BigUint::from(157u8));
                vec.push(BigUint::from(163u8));
                vec.push(BigUint::from(167u8));
                vec.push(BigUint::from(173u8));
                vec.push(BigUint::from(179u8));
                vec.push(BigUint::from(181u8));
                vec.push(BigUint::from(191u8));
                vec.push(BigUint::from(193u8));
                vec.push(BigUint::from(197u8));
                vec.push(BigUint::from(199u8));
                vec.push(BigUint::from(211u8));
                vec.push(BigUint::from(223u8));
                vec.push(BigUint::from(227u8));
                vec.push(BigUint::from(229u8));
                vec.push(BigUint::from(233u8));
                vec.push(BigUint::from(239u8));
                vec.push(BigUint::from(241u8));
                vec.push(BigUint::from(251u8)); //all primes that fit in a u8
                vec
            };

        }

        return is_prime(number, &DEFAULTVEC);
    }

    #[must_use]
    pub fn is_prime(number: &BigUint, g_primes: &Vec<BigUint>) -> bool {
        if BigUint::from(1u8) == *number {
            return false;
        }
        if BigUint::from(4u8) > *number {
            return true;
        }

        if number.sqrt().pow(2) == *number {
            return false;
        }

        let two = BigUint::from(2u8);

        // number = 2^a - 1
        // a = log2(number + 1)
        let a = log_2(&(number + 1u8));
        if BigUint::from(2u8).pow(a as u32) - BigUint::one() != *number {
            let mut i;
            let one = BigUint::one();
            let zero = BigUint::zero();

            let sqrtnum = number.sqrt() + &one; //fake ceil function

            for prime in g_primes {
                if prime < &sqrtnum && number % prime == zero {
                    return false;
                }
            }
            i = g_primes.iter().max().unwrap().clone();

            if !is_probably_prime(number, 5) {
                return false;
            }

            loop {
                i += &one;
                if number % &i == zero {
                    return false;
                }
                if i == sqrtnum {
                    return true;
                }
            }
        }

        // 4 12 194
        let mut last = BigUint::from(4u8);

        for _i in 2..a {
            last = (last.pow(2) - &two) % number;
        }

        last == BigUint::from(0u8)
    }
}
